/*
 * Project: irond
 * Package: src.de.fhhannover.inform.ifmap.datastructure
 * File:    Identifier2.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.ifmap.datastructure;

import java.util.Collections;
import java.util.Hashtable;
import java.util.Map;

import org.trustedcomputinggroup._2010.ifmap._2.AccessRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.DeviceType;
import org.trustedcomputinggroup._2010.ifmap._2.IPAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.IdentityType;
import org.trustedcomputinggroup._2010.ifmap._2.MACAddressType;

import de.fhhannover.inform.ifmap.exception.NoValidIdentifierTypeException;

/**
 * Represents an Identifier
 * 
 */
public class Identifier2 extends IfmapDataType {
	
	private final Hashtable<String, String> mAttributes = new Hashtable<String, String>();
	private final String mName;
	private final Object mRequestObject;

	private int mType;

	public Identifier2(Object o) throws NoValidIdentifierTypeException {
		super();
		if (o instanceof AccessRequestType) {
			mName = "access-request";
			AccessRequestType art = (AccessRequestType) o;
			mRequestObject = art;
			mAttributes.put("name", art.getName());
			if (art.getAdministrativeDomain() != null)
				mAttributes.put("administrative-domain",
						art.getAdministrativeDomain());
			this.mType = IdentifierConstants.ACCESS_REQUEST;
		} else if (o instanceof DeviceType) {
			mName = "device";
			DeviceType d = (DeviceType) o;
			mRequestObject = d;
			mAttributes.put("name", d.getName());
			if (d.getAikName() != null)
				mAttributes.put("aik-name", d.getAikName());
			this.mType = IdentifierConstants.DEVICE;
		} else if (o instanceof IPAddressType) {
			mName = "ip-address";
			IPAddressType ipt = (IPAddressType) o;
			mRequestObject = ipt;
			mAttributes.put("value", ipt.getValue());
			mAttributes.put("type", ipt.getType());
			if (ipt.getAdministrativeDomain() != null)
				mAttributes.put("administrative-domain",
						ipt.getAdministrativeDomain());
			this.mType = IdentifierConstants.IP_ADDRESS;
		} else if (o instanceof IdentityType) {
			mName = "identity";
			IdentityType it = (IdentityType) o;
			mRequestObject = it;
			mAttributes.put("name", it.getName());
			mAttributes.put("type", it.getType());
			if (it.getAdministrativeDomain() != null)
				mAttributes.put("administrative-domain",
						it.getAdministrativeDomain());
			this.mType = IdentifierConstants.IDENTITY;
		} else if (o instanceof MACAddressType) {
			mName = "mac-address";
			MACAddressType mt = (MACAddressType) o;
			mRequestObject = mt;
			mAttributes.put("value", mt.getValue());
			if (mt.getAdministrativeDomain() != null)
				mAttributes.put("administrative-domain",
						mt.getAdministrativeDomain());
			this.mType = IdentifierConstants.MAC_ADDRESS;
		} else {
			throw new NoValidIdentifierTypeException("Unknown Request Type: "
					+ o.getClass().getName());
		}
	}

	public int getType() {
		return mType;
	}

	public Object getRequestObject() {
		return mRequestObject;
	}	

	private String getAttributeValue(String key) {
		return mAttributes.get(key);
	}

	public Map<String, String> getAttributes() {
		// We return an unmodifiable Map
		return Collections.unmodifiableMap(mAttributes);
	}

	public String getName() {
		return mName;
	}

	@Override
	public boolean equals(Object o) {
		if (o == null)
			return false;
		if (this == o)
			return true;
		if (!(o instanceof Identifier2))
			return false;

		Identifier2 i2 = (Identifier2) o;
		
		if (mType != i2.getType())
			return false;

		if (this.hashCode() == i2.hashCode()) {
			return true;
		} else
			return false;
	}

	@Override
	public int hashCode() {
		StringBuffer buf = new StringBuffer();
		buf.append(this.getName());
		switch (mType) {
		case IdentifierConstants.ACCESS_REQUEST:
			buf.append("name=" + this.getAttributeValue("name"));
			buf.append("administrative-domain="
					+ this.getAttributeValue("administrative-domain"));
			break;
		case IdentifierConstants.IP_ADDRESS:
			buf.append("value=" + this.getAttributeValue("value"));
			buf.append("type=" + this.getAttributeValue("type"));
			break;
		case IdentifierConstants.DEVICE:
			buf.append("name=" + this.getAttributeValue("name"));
			buf.append("aik-name=" + this.getAttributeValue("aik-name"));
			break;
		case IdentifierConstants.IDENTITY:
			buf.append("name=" + this.getAttributeValue("name"));
			buf.append("type=" + this.getAttributeValue("type"));
			buf.append("administrative-domain="
					+ this.getAttributeValue("administrative-domain"));
			break;
		case IdentifierConstants.MAC_ADDRESS:
			buf.append("value=" + this.getAttributeValue("value"));
			buf.append("administrative-domain="
					+ this.getAttributeValue("administrative-domain"));
			break;
		}
		return buf.toString().hashCode();
	}
	
	@Override	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(this.getName());
		buf.append(mAttributes.toString()+ " {"+this.hashCode()+"} [M] "+
				this.getMetadata().toString());
		return buf.toString();
	}

}
