/*
 * Project: irond
 * Package: src.de.fhhannover.inform.ifmap.output
 * File:    JPrefuse.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.ifmap.output;

/**
 * The Implementation of a Prefuse-Graph
 */

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.table.DefaultTableModel;

import org.apache.log4j.Logger;

import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.action.ActionList;
import prefuse.action.RepaintAction;
import prefuse.action.assignment.ColorAction;
import prefuse.action.assignment.StrokeAction;
import prefuse.action.layout.graph.ForceDirectedLayout;
import prefuse.controls.ControlAdapter;
import prefuse.controls.FocusControl;
import prefuse.controls.NeighborHighlightControl;
import prefuse.controls.PanControl;
import prefuse.controls.SubtreeDragControl;
import prefuse.controls.ToolTipControl;
import prefuse.controls.WheelZoomControl;
import prefuse.controls.ZoomControl;
import prefuse.controls.ZoomToFitControl;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.expression.Predicate;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.render.DefaultRendererFactory;
import prefuse.render.EdgeRenderer;
import prefuse.render.LabelRenderer;
import prefuse.util.ColorLib;
import prefuse.util.force.DragForce;
import prefuse.util.force.ForceSimulator;
import prefuse.util.force.NBodyForce;
import prefuse.util.force.RungeKuttaIntegrator;
import prefuse.util.force.SpringForce;
import prefuse.visual.VisualItem;
import de.fhhannover.inform.ifmap.datastructure.Identifier2;
import de.fhhannover.inform.ifmap.datastructure.IdentifierContainer;
import de.fhhannover.inform.ifmap.datastructure.IfmapDataType;
import de.fhhannover.inform.ifmap.datastructure.Link2;
import de.fhhannover.inform.ifmap.datastructure.Metadata;
import de.fhhannover.inform.ifmap.datastructure.MetadataEntry;
import de.fhhannover.inform.ifmap.datastructure.MultiHashMap;
import de.fhhannover.inform.ifmap.event.IdentifierChangedReceiver;
import de.fhhannover.inform.ifmap.event.NodeSelectedReceiver;
import edu.emory.mathcs.backport.java.util.Collections;

@SuppressWarnings("serial")
public class JPrefuse extends Display implements ComponentListener,
		IdentifierChangedReceiver {

	private MultiHashMap<Identifier2, Node> identifier2node = new MultiHashMap<Identifier2, Node>();
	private MultiHashMap<Link2, Node> link2node = new MultiHashMap<Link2, Node>();
	private final Object mSemaphore = new Object();
	private Graph mGraph;
	private Visualization mVis;
	private LinkedHashSet<String> mPublishers = new LinkedHashSet<String>();
	private Color DEFAULT_IDENTIFIER_COLOR = new Color(153, 153, 255);
	private Color DEFAULT_METADATA_COLOR = new Color(255, 153, 102);
	private Predicate mPredicateIdentifier = ExpressionParser
			.predicate("type = '0'");
	private Predicate mPredicateMetadata = ExpressionParser
			.predicate("type = '1'");
	private JTable mTable;
	private HashMap<String, Color> mColorMap;
	private ArrayList<NodeSelectedReceiver> mReceiver = new ArrayList<NodeSelectedReceiver>();
	private String[] mMetatableColumns = new String[] { "Element",
			"Value / Attribute-Name", "Attribute-Value" };
	private boolean mAnimation = true;
	private Logger mLogger = Logger.getLogger(JPrefuse.class);
	private ActionList mLayout;
	private RepaintAction repaintAction;
	private ForceDirectedLayout forceLayoutAction;

	public JPrefuse() {
		setBackground(Color.WHITE);
		setHighQuality(true);
		init();
	}

	public void addNodeSelectedListener(NodeSelectedReceiver r) {
		if (!mReceiver.contains(r))
			mReceiver.add(r);
	}

	@Override
	public void componentHidden(ComponentEvent arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void componentMoved(ComponentEvent arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void componentResized(ComponentEvent arg0) {
		revalidate();
		pan(getParent().getWidth() / 2, getParent().getHeight() / 2);
	}

	@Override
	public void componentShown(ComponentEvent arg0) {
		// TODO Auto-generated method stub

	}

	public String[] getPublishers() {
		mPublishers.clear();
		for (Identifier2 ident : identifier2node.keys()) {
			for (Metadata meta : ident.getMetadata()) {
				mPublishers.add(meta.getPublisher());
			}
		}
		for (Link2 link : link2node.keys()) {
			for (Metadata meta : link.getMetadata()) {
				mPublishers.add(meta.getPublisher());
			}
		}

		if (mPublishers.size() > 0) {
			String[] pub = new String[mPublishers.size()];
			Iterator<String> it = mPublishers.iterator();
			int i = 0;
			while (it.hasNext()) {
				pub[i++] = it.next();
			}
			return pub;
		} else
			return null;
	}

	public void notifyNodeSelectedReceiver(VisualItem item) {
		for (NodeSelectedReceiver r : mReceiver) {
			r.nodeSelected(item);
		}
	}

	@Override
	public synchronized void processAll(ArrayList<IfmapDataType> newData,
			ArrayList<IfmapDataType> updateData,
			ArrayList<IfmapDataType> deleteData) {
		synchronized (mSemaphore) {
			this.stopActions();			
			if (newData != null) {
				for (IfmapDataType i : newData)
					this.addNode(i);
			}
			if (updateData != null) {
				for (IfmapDataType j : updateData)
					this.addNode(j);
			}
			if (deleteData != null) {
				for (IfmapDataType m : deleteData)
					this.removeNode(m);
			}				
			this.startActions();			
		}
	}

	@Override
	@Deprecated
	public void processDelete(IdentifierContainer conti) {

	}

	@Override
	@Deprecated
	public void processNew(IdentifierContainer conti) {
		
	}

	@Override
	@Deprecated
	public void processUpdate(IdentifierContainer conti) {

	}

	public void refreshColorMapping() {
		this.setNewMetadataColorAction(mColorMap, DEFAULT_IDENTIFIER_COLOR,
				DEFAULT_METADATA_COLOR);
	}

	public void setNewMetadataColorAction(HashMap<String, Color> map,
			Color iColor, Color mColor) {
		this.stopActions();
		ActionList fill = new ActionList();
		ColorAction fc1 = new ColorAction("graph.nodes", mPredicateIdentifier,
				VisualItem.FILLCOLOR, ColorLib.color(iColor));
		ColorAction fc2 = new ColorAction("graph.nodes", mPredicateMetadata,
				VisualItem.FILLCOLOR, ColorLib.color(mColor));
		fill.add(fc1);
		fill.add(fc2);
		if (map != null && map.size() > 0) {
			mColorMap = map;
			ArrayList<String> arr = new ArrayList<String>();
			arr.addAll(map.keySet());
			Collections.sort(arr);
			int[] palette = new int[arr.size()];
			int i = 0;
			for (String key : arr) {
				Color c = map.get(key);
				palette[i] = ColorLib.color(c);
				Predicate predicate = ExpressionParser
						.predicate("publisher = '" + key + "'");
				ColorAction fillColor = new ColorAction("graph.nodes",
						predicate, VisualItem.FILLCOLOR, palette[i]);
				fill.add(fillColor);
				i++;
			}
		}
		mVis.putAction("fill", fill);
		this.DEFAULT_IDENTIFIER_COLOR = iColor;
		this.DEFAULT_METADATA_COLOR = mColor;
		this.startActions();
	}

	public void setTable(JTable tab, JLabel label) {
		this.mTable = tab;
		// this.mLabel = label;
	}

	public synchronized void toggleAnimation() {
		if(mAnimation){
			// disable it
			mVis.cancel("layout");
		} else {
			// enable it
			mVis.run("layout");
		}
		// update flag
		mAnimation = !mAnimation;
	}

	private void addIdentifier(Identifier2 identifier) {
		Node iNode = null;
		ArrayList<Metadata> metaListNew = identifier.getMetadata();
		if (!identifier2node.hasKey(identifier)) {
			Node n = mGraph.addNode();
			setIdentifierNodeValue(n, identifier);
			identifier2node.put(identifier, n);
			iNode = n;
			for (Metadata meta : metaListNew) {
				Node m = mGraph.addNode();
				setMetadataNodeValue(m, meta);
				mGraph.addEdge(m, iNode);
				identifier2node.put(identifier, m);
				this.mPublishers.add(meta.getPublisher());
			}
		} else {
			iNode = identifier2node.get(identifier).get(0);
			ArrayList<Metadata> metaListOld = identifier2node
					.getKey(identifier).getMetadata();
			if (metaListNew != null && !metaListNew.isEmpty()) {
				Identifier2 oldIdent = (Identifier2) identifier2node
						.getKey(identifier);
				for (Metadata meta : metaListNew) {
					if (!metaListOld.contains(meta)) {
						Node m = mGraph.addNode();
						setMetadataNodeValue(m, meta);
						mGraph.addEdge(m, iNode);
						identifier2node.put(identifier, m);
						this.mPublishers.add(meta.getPublisher());
						if (oldIdent != null) {
							oldIdent.addOrReplaceMetadata(meta);
						}
					} else {
						Metadata mOld = metaListOld.get(metaListOld
								.indexOf(meta));
						if (mOld.getTimestamp().equalsIgnoreCase(
								meta.getTimestamp())) {
							return;
						}
						if (!meta.isSingleValue()) {
							Node m = mGraph.addNode();
							setMetadataNodeValue(m, meta);
							mGraph.addEdge(m, iNode);
							identifier2node.put(identifier, m);
							if (oldIdent != null) {
								oldIdent.addOrReplaceMetadata(meta);
							}
						}
					}

				}
			}
		}
	}

	private void addLink(Link2 link) {
		// first we create the identifiers if they not exists
		Identifier2 i1 = link.getIdentifier1();
		Identifier2 i2 = link.getIdentifier2();
		if (!identifier2node.hasKey(i1)) {
			Node m = mGraph.addNode();
			setIdentifierNodeValue(m, i1);
			identifier2node.put(i1, m);
		}
		if (!identifier2node.hasKey(i2)) {
			Node n = mGraph.addNode();
			setIdentifierNodeValue(n, i2);
			identifier2node.put(i2, n);
		}
		ArrayList<Metadata> metaData = link.getMetadata();
		if (!link2node.hasKey(link)) {
			for (Metadata meta : metaData) {
				Node l = mGraph.addNode();
				setMetadataNodeValue(l, meta);

				// add edges
				mGraph.addEdge(identifier2node.get(i1).get(0), l);
				mGraph.addEdge(identifier2node.get(i2).get(0), l);

				link2node.put(link, l);
				this.mPublishers.add(meta.getPublisher());
			}
		} else {
			Link2 oldLink = (Link2) link2node.getKey(link);
			if (oldLink != null) {
				for (Metadata meta : link.getMetadata()) {
					oldLink.addOrReplaceMetadata(meta);
				}
				ArrayList<Node> nodeList = link2node.get(oldLink);
				for (Node node : nodeList) {
					mGraph.removeNode(node);
				}
				nodeList.clear();
				for (Metadata meta : oldLink.getMetadata()) {
					Node l = mGraph.addNode();
					setMetadataNodeValue(l, meta);
					// add edges
					mGraph.addEdge(identifier2node.get(i1).get(0), l);
					mGraph.addEdge(identifier2node.get(i2).get(0), l);
					nodeList.add(l);
				}
			}
		}

	}

	private void addNode(IfmapDataType ifmap) {
		if (ifmap instanceof Identifier2) {
			addIdentifier((Identifier2) ifmap);
		} else if (ifmap instanceof Link2) {
			addLink((Link2) ifmap);
		}
	}

//	private void enableAnimation(boolean b) {
//		if (b) {
//			if (!mVis.getAction("layout").isRunning()) {
//				mVis.run("layout");
//			}
//		} else {
//			mVis.cancel("layout");
//		}
//		mAnimation = b;
//	}

	private boolean hasLinks(Identifier2 ident) {
		boolean hasLinks = false;
		for (Link2 link : this.link2node.keys()) {
			if (link.getIdentifier1().equals(ident)
					|| link.getIdentifier2().equals(ident)) {
				hasLinks = true;
			}
		}
		return hasLinks;
	}

	private void init() {
		if (mGraph == null) {
			mGraph = new Graph(false);
			mGraph.getNodeTable().addColumn("name", String.class);
			mGraph.getNodeTable().addColumn("value", String.class);
			mGraph.getNodeTable().addColumn("type", String.class);
			mGraph.getNodeTable().addColumn("publisher", String.class);
			mGraph.getNodeTable().addColumn("desc", String.class);
			mGraph.getNodeTable().addColumn("object", Object.class);
			mGraph.getNodeTable().addColumn(VisualItem.SHAPE, Integer.class);
		}

		// nodes
		LabelRenderer labelRend = new LabelRenderer("value");
		labelRend.setRoundedCorner(4, 4);
		labelRend.setHorizontalPadding(4);
		labelRend.setVerticalPadding(4);

		// edges
		EdgeRenderer edgeRend = new EdgeRenderer(Constants.EDGE_TYPE_CURVE);

		// create a new Visualisation
		mVis = new Visualization();
		mVis.add("graph", mGraph);
		mVis.setRendererFactory(new DefaultRendererFactory(labelRend, edgeRend));

		// set all identifiers to the same color
		ColorAction borderColor = new ColorAction("graph.nodes",
				VisualItem.STROKECOLOR, ColorLib.rgb(0, 0, 0));
		ColorAction textColor = new ColorAction("graph.nodes",
				VisualItem.TEXTCOLOR, ColorLib.rgb(0, 0, 0));
		ColorAction edgeColor = new ColorAction("graph.edges",
				VisualItem.STROKECOLOR, ColorLib.rgb(0, 0, 0));
		ColorAction arrowColor = new ColorAction("graph.edges",
				VisualItem.FILLCOLOR, ColorLib.rgb(0, 0, 0));
		ColorAction fillColor = new ColorAction("graph.nodes",
				mPredicateIdentifier, VisualItem.FILLCOLOR,
				ColorLib.color(DEFAULT_IDENTIFIER_COLOR));
		ColorAction fillColor2 = new ColorAction("graph.nodes",
				mPredicateMetadata, VisualItem.FILLCOLOR,
				ColorLib.color(DEFAULT_METADATA_COLOR));
		ActionList fill = new ActionList();
		fill.add(fillColor);
		fill.add(fillColor2);
		mVis.putAction("fill", fill);

		ActionList color = new ActionList();
		color.add(borderColor);
		color.add(textColor);
		color.add(edgeColor);
		color.add(arrowColor);

		// the stroke
		ActionList stroke = new ActionList();
		stroke.add(new StrokeAction("graph.edges", new BasicStroke(2)));

		mLayout = new ActionList(ActionList.INFINITY);
		forceLayoutAction = new ForceDirectedLayout("graph", false);
		ForceSimulator fsim = new ForceSimulator(new RungeKuttaIntegrator());
		float gravConstant = -40f; // the more negative, the more repelling
		float minDistance = 300f; // -1 for always on, the more positive, the
									// more space between nodes
		float theta = 0.3f; // the lower, the more single-node repell
							// calculation
		float drag = 0.01f;
		float springCoeff = 1E-4f; // 1E-4
		fsim.addForce(new NBodyForce(gravConstant, minDistance, theta));
		fsim.addForce(new DragForce(drag));
		fsim.addForce(new SpringForce(springCoeff, 5));
		fsim.getForces()[2].setParameter(SpringForce.SPRING_LENGTH, 150f);
		forceLayoutAction.setForceSimulator(fsim);

		repaintAction = new RepaintAction();

		mLayout.add(0, forceLayoutAction);
		mLayout.add(1, repaintAction);

		// putting it all in the Visualisation
		mVis.putAction("color", color);
		mVis.putAction("stroke", stroke);
		mVis.putAction("layout", mLayout);
		// mVis.putAction("shape", mLayout);
		mVis.putAction("fill", fill);

		// create display and controls
		this.setVisualization(mVis);
		this.addControlListener(new SubtreeDragControl());
		this.addControlListener(new PanControl());
		this.addControlListener(new ZoomControl());
		this.addControlListener(new ZoomToFitControl());
		this.addControlListener(new FocusControl(2, "layout"));
		this.addControlListener(new WheelZoomControl());
		this.addControlListener(new NeighborHighlightControl());

		// listener for node selection
		this.addControlListener(new ControlAdapter() {
			private int oldColor;
			private Font oldFont;

			public void itemPressed(VisualItem item, MouseEvent e) {
				if (item.canGet("object", Object.class)) {
					oldColor = item.getFillColor();
					oldFont = item.getFont();
					item.setFillColor(ColorLib.color(Color.RED));
					item.setFont(oldFont.deriveFont(Font.BOLD));
					showMetaInTable(item);
					notifyNodeSelectedReceiver(item);
				}
			}

			public void itemReleased(VisualItem item, MouseEvent e) {
				item.setFillColor(oldColor);
				item.setFont(oldFont);
			}
		});

		addControlListener(new ToolTipControl("name") {
			{
				javax.swing.ToolTipManager.sharedInstance().setDismissDelay(
						60000);
				javax.swing.ToolTipManager.sharedInstance().setInitialDelay(0);
			}

			public void itemEntered(VisualItem item, java.awt.event.MouseEvent e) {
				Display d = (Display) e.getComponent();
				if (item.canGet("object", Object.class)) {
					StringBuffer toolTipText = new StringBuffer();
					toolTipText.append("<html>");
					if (item.get("object") instanceof Metadata) {
						Metadata meta = (Metadata) item.get("object");
						toolTipText.append("<strong><u>" + meta.getName()
								+ "</u></strong><br>");
						toolTipText.append("<i>ifmap-publisher-id:</i> "
								+ meta.getPublisher() + "<br>");
						toolTipText.append("<i>ifmap-timestamp:</i> "
								+ meta.getTimestamp() + "<br>");
						toolTipText.append("<i>ifmap-cardinality:</i> "
								+ meta.getCardinality() + "<br>");
						if (meta.hasAttributes()) {
							toolTipText
									.append("<strong>Attributes</strong><br>");
							for (String key : meta.getAttributes().keySet()) {
								toolTipText.append("<i>" + key + ":</i> "
										+ meta.getAttribute(key) + "<br>");
							}
						}
						if (meta.hasElements()) {
							toolTipText.append("<strong>Elements</strong><br>");
							for (MetadataEntry entry : meta.getElements()) {
								if (!entry.hasAttributes()) {
									toolTipText.append("<i>" + entry.getName()
											+ ":</i> " + entry.getValue()
											+ "<br>");
								} else {
									toolTipText.append("<i>" + entry.getName()
											+ ":</i><br>");
									HashMap<String, String> attr = entry
											.getAttributes();
									for (String key : attr.keySet()) {
										toolTipText
												.append("<span>&#160;&#160;&#160;</span>");
										toolTipText.append("<i>" + key
												+ ":</i> " + attr.get(key));
										toolTipText.append("<br>");
									}
								}
							}
						}
					} else if (item.get("object") instanceof Identifier2) {
						Identifier2 ident = (Identifier2) item.get("object");
						toolTipText.append("<strong><u>" + ident.getName()
								+ "</u></strong><br>");
						Map<String, String> attr = ident.getAttributes();
						if (attr != null && attr.size() > 0) {
							for (String key : attr.keySet()) {
								toolTipText.append("<i>" + key + ":</i> "
										+ attr.get(key) + "<br>");
							}
						}
					}
					toolTipText.append("</html>");
					d.setToolTipText(toolTipText.toString());
				}
			}
		});

		this.zoom(getLocation(), 0.9);
		this.startActions();
	}

	private void removeIdentifier(Identifier2 identifier) {
		if (identifier2node.hasKey(identifier)) {
			Identifier2 oldIdent = (Identifier2) identifier2node
					.getKey(identifier);
			if (oldIdent != null) {
				ArrayList<Node> nodeList = identifier2node.get(oldIdent);
				ArrayList<Metadata> metaListOld = oldIdent.getMetadata();
				ArrayList<Metadata> metaListNew = identifier.getMetadata();
				for (int i = 0; i < metaListNew.size(); i++) {
					Metadata meta = metaListNew.get(i);
					if (metaListOld.contains(meta)) {
						mLogger.trace("removing metadata: " + meta);
						// find node that matches meta object
						for (int j = 0; j < nodeList.size(); j++) {
							Object tmp = nodeList.get(j).get("object");
							// check if meta object matches the one to be
							// removed
							if (tmp instanceof Metadata && tmp.equals(meta)) {
								mGraph.removeNode(nodeList.get(j));
								metaListOld.remove(tmp);
								nodeList.remove(j);
							}
						}
					}
				}
				if (nodeList.size() == 1) {
					if (!hasLinks(identifier)) {
						mLogger.trace("identifier has no metadata left, removing it ... "
								+ identifier);
						mGraph.removeNode(nodeList.get(0));
						nodeList.remove(0);
						identifier2node.remove(identifier);
					}
				}
			}
		}

	}

	private void removeLink(Link2 link) {
		if (link2node.hasKey(link)) {
			Link2 oldLink = (Link2) link2node.getKey(link);
			if (oldLink != null) {
				ArrayList<Node> nodeList = link2node.get(oldLink);
				ArrayList<Metadata> metaListOld = oldLink.getMetadata();
				ArrayList<Metadata> metaListNew = link.getMetadata();
				for (int i = 0; i < metaListNew.size(); i++) {
					Metadata meta = metaListNew.get(i);
					if (metaListOld.contains(meta)) {
						for (int j = 0; j < nodeList.size(); j++) {
							Node node = nodeList.get(j);
							Metadata metaOld = (Metadata) node.get("object");
							if (metaOld.equals(meta)) {
								mLogger.trace("removing metadata from link: "
										+ meta);
								nodeList.remove(node);
								mGraph.removeNode(node);
								metaListOld.remove(meta);
							}
						}
					}
					if (nodeList.size() == 0) {
						mLogger.trace("link has no metadata left, removing it ... "
								+ link);
						link2node.remove(link);
						this.removeIdentifier(oldLink.getIdentifier1());
						this.removeIdentifier(oldLink.getIdentifier2());
					}
				}
			}
		}
	}

	private void removeNode(IfmapDataType ifmap) {
		if (ifmap instanceof Identifier2) {
			removeIdentifier((Identifier2) ifmap);
		} else if (ifmap instanceof Link2) {
			removeLink((Link2) ifmap);
		}
	}

	private void setIdentifierNodeValue(Node node, Identifier2 ident) {
		node.setString("name", ident.getName());
		StringBuffer buf = new StringBuffer();
		Map<String, String> map = ident.getAttributes();
		buf.append(ident.getName());
		buf.append("\r\n");
		for (String key : ident.getAttributes().keySet()) {
			buf.append("[ ");
			buf.append(key);
			buf.append("=");
			buf.append(map.get(key));
			buf.append(" ]");
			buf.append("\r\n");
		}
		buf.delete(buf.length() - 1, buf.length());
		node.setString("value", buf.toString());
		node.setString("publisher", "");
		node.setString("type", "0");
		node.setString("desc", ident.getName() + ","
				+ ident.getAttributes().toString());
		node.set("object", ident);
	}

	private void setMetadataNodeValue(Node node, Metadata meta) {
		node.setString("name", meta.getName());
		node.setString("value", meta.getName());
		node.setString("type", "1");
		node.setString("publisher", meta.getPublisher());
		node.setString("desc", meta.toString());
		node.set("object", meta);
	}

	private void showMetaInTable(VisualItem item) {
		if (mTable != null) {
			if (item.canGet("object", Object.class)) {
				DefaultTableModel model = new DefaultTableModel();
				for (String s : mMetatableColumns) {
					model.addColumn(s);
				}
				if (item.get("object") instanceof Metadata) {
					Metadata meta = (Metadata) item.get("object");
					ArrayList<MetadataEntry> values = meta.getElements();
					if (meta.hasAttributes()) {
						for (String key : meta.getAttributes().keySet()) {
							String[] row = {
									"<html><strong>" + key + "</strong></html>",
									meta.getAttribute(key), "" };
							model.addRow(row);
						}
					}
					for (MetadataEntry entry : values) {
						if (!entry.hasAttributes()) {
							String[] row = {
									"<html><strong>" + entry.getName()
											+ "</strong></html>",
									entry.getValue(), "" };
							model.addRow(row);
						} else {
							String[] head = {
									"<html><strong>" + entry.getName()
											+ "</strong></html>", "", "" };
							model.addRow(head);
							HashMap<String, String> attr = entry
									.getAttributes();
							for (String key : attr.keySet()) {
								String[] row = { "", key, attr.get(key) };
								model.addRow(row);
							}
						}
					}
				} else if (item.get("object") instanceof Identifier2) {
					Identifier2 ident = (Identifier2) item.get("object");
					Map<String, String> attributes = ident.getAttributes();
					if (attributes != null) {
						Set<String> keys = attributes.keySet();
						if (keys != null) {
							for (String key : keys) {
								String[] vrow = {
										"<html><strong>" + key
												+ "</strong></html>",
										attributes.get(key), "" };
								model.addRow(vrow);
							}
						}
					}
				}
				mTable.setModel(model);
			}
		}
	}

	private synchronized void startActions() {
		mVis.run("fill");
		mVis.run("color");
		mVis.run("stroke");		
		mVis.run("layout");
//		if (!mPublishers.isEmpty()) {
//			for (String key : mPublishers) {
//				mVis.run(key);
//			}
//		}
		
	}

	private synchronized void stopActions() {
		mVis.cancel("fill");
		mVis.cancel("color");
		mVis.cancel("stroke");
		mVis.cancel("layout");		
		if (!mPublishers.isEmpty()) {
			for (String key : mPublishers) {
				mVis.cancel(key);
			}
		}
	}
}